#############################################################
###                                                       ###
### This program reproduces all of the results in the MST ###
### manuscript.  However, the tooth analysis takes around ###
### 30 min to run. One may load the tooth analysis trees  ###
### if preferable.                                        ###
###                                                       ###
#############################################################


############################################################
#####            Load and Create Functions             #####
#####                                                  #####
############################################################

## Load MST package after installing it from CRAN or local source.
library("MST") 
library("partykit")
library("microbenchmark")
library("plyr")
## Source modified MST partykit functions (may need to set
## working directory).
source("partykit-functions.R")

as.numeric.factor <- function(x) {
    as.numeric(levels(x))[x]
}

if (!dir.exists("Figures")) {
    dir.create("Figures")
}

############################################################
#####            Generate Simulated Data               #####
#####                                                  #####
############################################################

set.seed(186117)
data <- rmultime(N = 200, K = 4, beta = c(-1, 0.8, 0.8, 0, 0),
                 cutoff = c(0.5, 0.3, 0, 0), model = "marginal.multivariate.exponential",
                 rho = 0.65)$dat
test <- rmultime(N = 100, K = 4, beta = c(-1, 0.8, 0.8, 0, 0),
                 cutoff = c(0.5, 0.3, 0, 0), model = "marginal.multivariate.exponential",
                 rho = 0.65)$dat

############################################################
#####      Analyze Simulated Data and Plot Trees       #####
#####                                                  #####
############################################################

fit <- MST(formula = Surv(time, status) ~ x1 + x2 + x3 + x4 | id,
           data, test, method = "marginal", minsplit = 100, minevents = 20,
           selection.method = "test.sample")

pdf(file = "Figures/MST_simulation_panelA.pdf",
  width = 15.97, height = 8.33, pointsize = 12, bg = "white")
tree_initial <- getTree(fit, "0")
plot(tree_initial, type = "simple", main = " ",
     inner_panel = node_inner_fig1(tree_initial), edge_panel = edge_simple_fig1(tree_initial))
grid.text("Initial Tree", x = 0.5, y = 0.95, gp = gpar(fontsize = 48, font = 2))
grid.text("A", x = 0.05, y = 0.962, gp = gpar(fontsize = 64, font = 2))
dev.off()

pdf(file = "Figures/MST_simulation_panelB.pdf",
  width = 6.94, height = 2.78, pointsize = 12, bg = "white")
par(mar = c(2, 2, 1.7, 0))
plot(NULL, xlim = c(1, 14), ylim = c(-30, 50), xlab = "", ylab = "", axes = FALSE)
text(x = 0.25, y = 61, "B", xpd = TRUE, cex = 2.5, font = 2)
title("Selecting Best Tree Size", cex.main = 1.8, line = 0)
lines(as.numeric.factor(fit[["pruning.info"]][["size.tmnl"]]), as.numeric.factor(fit[["pruning.info"]][["Ga.2"]]), lty = 1, lwd = 2)
lines(as.numeric.factor(fit[["pruning.info"]][["size.tmnl"]]), as.numeric.factor(fit[["pruning.info"]][["Ga.3"]]), lty = 2, lwd = 2)
lines(as.numeric.factor(fit[["pruning.info"]][["size.tmnl"]]), as.numeric.factor(fit[["pruning.info"]][["Ga.4"]]), lty = 3, lwd = 2)
lines(as.numeric.factor(fit[["pruning.info"]][["size.tmnl"]]), as.numeric.factor(fit[["pruning.info"]][["Ga.log_n"]]), lty = 4, lwd = 2)
mtext("Goodness of Split", 2, cex = 1.1, font = 2, line = 1)
mtext("# of Terminal Nodes", 1, cex = 1.1, font = 2, line = 1)
axis(side = 1, at = 1:14, font = 2, cex.axis = 1, padj = -1, xpd = TRUE, line = -0.2)
axis(side = 2, at = seq(-25, 50, 25), las = 2, font = 2, cex.axis = 1, hadj = .75, line = -1)
legend(3, 15, c(expression(alpha == 2), expression(alpha == 3), expression(alpha == 4), expression(alpha == log(n[0]))), lty = c(1, 2, 3, 4), lwd  =  2, bty  =  "n", cex  =  1)
dev.off()

pdf(file = "Figures/MST_simulation_panelC.pdf",
  width = 6.94, height = 3.47, pointsize = 12, bg = "white")
(tree_final <- getTree(fit, "4"))
plot(tree_final, main = " ")
grid.text("Final Tree", x = .5, y = .94, gp = gpar(fontsize = 24, font = 2))
grid.text("C", x = .05, y = .95, gp = gpar(fontsize = 32, font = 2))
dev.off()

## Fit a model from the final tree.
data$term_nodes <- as.factor(predict(tree_final, newdata = data, type = "node"))
coxph(Surv(time, status) ~ term_nodes + cluster(id), data = data)

############################################################
#####             Analysis of Tooth Loss               #####
#####                                                  #####
############################################################

data("Teeth", package = "MST")

## Explore Teeth data.
dim(Teeth)  # Number of teeth 
subjects <- Teeth[!duplicated(Teeth$id), c("id", paste0("x", 48:51))]
dim(subjects)  # Number of patients 
table(Teeth$molar)  # Number of molars/non-molars 
mean(subjects$x48)  # Average age 
prop.table(table(subjects$x49))
prop.table(table(subjects$x50))
prop.table(table(subjects$x51))

## Molars Analysis 
molar <- subset(Teeth, molar == 1); dim(molar)
n <- NROW(molar)  # Number of molars 
nPatients <- length(unique(molar$id))  # Number of patients with molar teeth 

## Explore molars data.
sum(molar$event == 1, na.rm = TRUE); mean(molar$event == 1, na.rm = TRUE)  # Number (%) of molars lost
median(molar$time[molar$event == 1], na.rm = TRUE)  # Median molar loss time
max(molar$time[molar$event == 1], na.rm = TRUE)  # Maximum molar loss time

## Split dataset into training and test sample.
set.seed(482046)
## Split 2/3 patients (not teeth) into training dataset.
id.train <- sample(x = unique(molar$id), size = floor(nPatients * 2/3), replace = FALSE)
molar_training <- molar[molar$id %in% id.train, ]
molar_test <- molar[!(molar$id %in% id.train), ]

## Fit Molar Model (takes approx 15min).  Can load saved data (if  available).
if ("fit_molar.RData" %in% list.files()) {
    load("fit_molar.RData")
} else {
    form <- as.formula(paste0("Surv(time, event) ~ ", paste0("x", 1:51, collapse = " + "), " | id"))
    fit_molar <- MST(form, data = molar_training, test = molar_test,
                     method = "exp.frailty", selection.method = "test.sample",
                     distinct = FALSE, delta = 0.05, nCutPoints = 50, minevents=10,
                     minbucket = 100)
    save(fit_molar, file = "fit_molar.RData")
}

treeGa4_molar <- getTree(fit_molar, "4")
## Number of factors in final molar model.
unique(unlist(nodeapply(treeGa4_molar, ids = nodeids(treeGa4_molar)[-nodeids(treeGa4_molar, terminal = TRUE)], FUN = function(n) split_node(n)$varid)))

## Categorize each tooth into one of the terminal nodes.
molar_training$term_nodes <- as.factor(predict(treeGa4_molar, newdata = molar_training, type = "node"))
table(molar_training$term_nodes)  # Number of teeth in each node (will match size from tree)
## Calculate estimated survival at 3 years for each terminal node.
mod_molar <- coxph(Surv(time, event) ~ term_nodes + cluster(id), data = molar_training)
coef(mod_molar)
newdata_molar <- data.frame(term_nodes = levels(molar_training$term_nodes),
                            row.names = levels(molar_training$term_nodes))
temp <- survfit(mod_molar, newdata = newdata_molar)
surv3yr_molar <- apply(temp$surv, 2, function(x) { x[temp$time >= 3][1] })  # Estimated % of survival at 3 years
c(max(surv3yr_molar), min(surv3yr_molar))  # Survival rate range for molars

## Non-Molars Analysis

nonmolar <- subset(Teeth, molar == 0); dim(nonmolar)
n <- NROW(nonmolar)  # Number of non-molars 
nPatients <- length(unique(nonmolar$id))  # Number of patients 

## Explore non-molars dat.
sum(nonmolar$event == 1, na.rm = TRUE); mean(nonmolar$event == 1, na.rm = TRUE)  # Number (%) of non-molars lost
median(nonmolar$time[nonmolar$event == 1], na.rm = TRUE)  # Median non-molar loss time
max(nonmolar$time[nonmolar$event == 1], na.rm = TRUE)  # Maximum non-molar loss time

## Split dataset into training and test sample.
set.seed(482046)  # Use same seed as before to make code easier to replicate for users
## Split 2/3 patients (not teeth) into training dataset.
id.train <- sample(x = unique(nonmolar$id), size = floor(nPatients * 2/3), replace = FALSE)
nonmolar_training <- nonmolar[nonmolar$id %in% id.train, ]
nonmolar_test <- nonmolar[!(nonmolar$id %in% id.train), ]

## Fit Non-molar Model (takes approx 15min).  Can load saved data (if available).
if ("fit_nonmolar.RData" %in% list.files()) {
    load("fit_nonmolar.RData")
} else {
    form <- as.formula(paste0("Surv(time, event) ~ ", paste0("x", c(1:21, 23:51), collapse = " + "), " | id"))
    fit_nonmolar <- MST(form, data = nonmolar_training, test = nonmolar_test,
                        method = "exp.frailty", selection.method = "test.sample",
                        distinct = FALSE, delta = 0.05, nCutPoints = 50, minevents=10,
                        minbucket = 100)
    save(fit_nonmolar, file = "fit_nonmolar.RData")
}

treeGa4_nonmolar <- getTree(fit_nonmolar, "4")
## Number of factors in final non-molar model.
unique(unlist(nodeapply(treeGa4_nonmolar, ids = nodeids(treeGa4_nonmolar)[-nodeids(treeGa4_nonmolar, terminal = TRUE)], FUN = function(n) { split_node(n)$varid })))

## Categorize each tooth into one of the terminal nodes.
nonmolar_training$term_nodes <- as.factor(predict(treeGa4_nonmolar, newdata = nonmolar_training, type = "node"))
table(nonmolar_training$term_nodes)  # Number of teeth in each node (will match size from tree)
## Calculate estimated survival at 3 years for each terminal node.
mod_nonmolar <- coxph(Surv(time, event) ~ term_nodes + cluster(id), data = nonmolar_training)
coef(mod_nonmolar)
newdata_nonmolar <- data.frame(term_nodes = levels(nonmolar_training$term_nodes),
                               row.names = levels(nonmolar_training$term_nodes))
temp <- survfit(mod_nonmolar, newdata = newdata_nonmolar)
surv3yr_nonmolar <- apply(temp$surv, 2, function(x) { x[temp$time >= 3][1] })  # Estimated % of survival at 3 years
c(max(surv3yr_nonmolar), min(surv3yr_nonmolar))  # Survival rate range for molars


############################################################
#####             Plot Tooth Loss Trees                #####
#####                                                  #####
############################################################

extractCutpt <- function(x) {
  if(is.null(split_node(x)$breaks)) {
      out <- ""
  } else if(split_node(x)$index[1] == 1) {
      out <- paste0("\u2264", format(round(split_node(x)$breaks, 2), digits = 3))
  } else {
      out <- paste0(">", format(round(split_node(x)$breaks,2), digits = 3))
  }
  return(out)
}

ns <- table(predict(treeGa4_molar, type = "node"))
cutpts <- unlist(nodeapply(treeGa4_molar,
                           ids = nodeids(treeGa4_molar)[-nodeids(treeGa4_molar, terminal = TRUE)],
                           FUN = extractCutpt))

pdf(file = "Figures/MST_molars.pdf",
  width = 12.5, height = 8,pointsize = 12, bg = "white")
plot(treeGa4_molar, type = "simple", main = "",
     inner_panel = node_inner_fig2(treeGa4_molar, cutpts),
     edge_panel = edge_simple_fig2_3(treeGa4_molar),
     terminal_panel = node_terminal_fig2(treeGa4_molar, ns = ns,
       surv = format(surv3yr_molar, digits = 1, nsmall = 2)))

## Easiest to manually add categorical splits in tree.
grid.text(expression(phantom(x) %in% "{Not"), x = .243, y = .718, just = "left", gp = gpar(fontsize = 10, font = 2))
grid.text(expression("Decayed}"), x = .249, y = .702, just = "left", gp = gpar(fontsize = 10, font = 2))

grid.text(expression(phantom(x) %in% "{Not"), x = .437, y = .554, just = "left", gp = gpar(fontsize = 10, font = 2))
grid.text(expression("Decayed}"), x = .442, y = .539, just = "left", gp = gpar(fontsize = 10, font = 2))

grid.text(expression(phantom(x) %in% "{No Endo"), x = .141, y = .472, just = "left", gp = gpar(fontsize = 10, font = 2))
grid.text(expression("Therapy}"), x = .153, y = .457, just = "left", gp = gpar(fontsize = 10, font = 2))

grid.text(expression(phantom(x) %in% "{No Endo"), x = .368, y = .308, just = "left", gp = gpar(fontsize = 10, font = 2))
grid.text(expression("Therapy}"), x = .38, y = .293, just = "left", gp = gpar(fontsize = 10, font = 2))
dev.off()


ns <- table(predict(treeGa4_nonmolar, type = "node"))
cutpts <- unlist(nodeapply(treeGa4_nonmolar,
                           ids = nodeids(treeGa4_nonmolar)[-nodeids(treeGa4_nonmolar, terminal = TRUE)],
                           FUN = extractCutpt))

pdf(file = "Figures/MST_nonmolars.pdf",
   width = 12.5, height = 8, pointsize = 12, bg = "white")
plot(treeGa4_nonmolar, type = "simple", main = "",
     inner_panel = node_inner_fig3(treeGa4_nonmolar, cutpts),
     edge_panel = edge_simple_fig2_3(treeGa4_nonmolar),
     terminal_panel = node_terminal_fig3(treeGa4_nonmolar, ns = ns,
                                         surv = format(surv3yr_nonmolar, digits = 1, nsmall = 2)))

## Easiest to manually add categorical splits in tree.
grid.text(expression(phantom(x) %in% "{No Endo"), x = .140, y = .61, just = "left", gp = gpar(fontsize = 9, font = 2))
grid.text(expression("Therapy}"), x = .152, y = .595, just = "left", gp = gpar(fontsize = 9, font = 2))

grid.text(expression(phantom(x) %in% "{Not"), x = .125, y = .431, just = "left", gp = gpar(fontsize = 9, font = 2))
grid.text(expression("Decayed}"), x = .13, y = .417, just = "left", gp = gpar(fontsize = 9, font = 2))

dev.off()

############################################################
#####          Computation Time Comparison             #####
#####                                                  #####
############################################################

## This function converts continuous predictor to nominal.
contToCat <- function(var) {
    as.character(LETTERS[cut(var, seq(0, 1, .125) + c(-0.00001, rep(0, 7), 0.00001), labels = FALSE)])
}

## This function simulates a single dataset and returns the estimated
## time to construct a tree.
comp_time <- function(N, K, nVars, pctCat, method, seed = 861504) {
  set.seed(seed)
  data <- rmultime(N = N, K = K, beta = c(-1, 0, 1, 0, 1, rep(0, nVars - 4)),
                   cutoff = rep(0.5, nVars), model = "marginal.multivariate.exponential",
                   rho = 0.65, digits = 2)$dat
  test <- rmultime(N = N / 2, K = K, beta = c(-1, 0, 1, 0, 1, rep(0, nVars - 4)),
                   cutoff = rep(0.5, nVars), model = "marginal.multivariate.exponential",
                   rho = 0.65, digits = 2)$dat
  ## Rearrange datasets so first half of variables are all patient-level.
  data <- data[, c(1:4, seq(6, 4 + nVars, 2), seq(5, 4 + nVars, 2))]
  test <- test[, c(1:4, seq(6, 4 + nVars, 2), seq(5, 4 + nVars, 2))]
  ## Convert half the patient-level and night-level variables to categorical.
  ## Note: there are two informative variables (x2 and x4).  One will be continuous, the other nominal.
  varsCat <- c(6:(5 + (pctCat * nVars / 2)), (5 + nVars / 2):(4 + nVars / 2 + (pctCat * nVars / 2)))
  for (i in varsCat) {
    data[[names(data)[i]]] <- contToCat(data[[names(data)[i]]])
    test[[names(test)[i]]] <- contToCat(test[[names(test)[i]]])
  }
  
  form <- as.formula(paste0("Surv(time, status) ~ ", paste(names(data)[5:ncol(data)], collapse = " + "), " | id"))
  ## Computation time to get final tree.
  time_hr <- system.time(MST(form, data, test, method = method, selection.method = "test.sample", 
                             minsplit = 20, minevents = 3, minbucket = 5, maxdepth = 10,
                             sortTrees = FALSE, plot.Ga = FALSE))
  return(time_hr / 60)
}

## Use microbenchmark to run multiple simulation times.

## Note: The comp_time function takes a long time to run.  It is recommended to simply load the results (if available).

## Sample size 800 
if ("comp_results_800.RData" %in% list.files()) {
    load("comp_results_800.RData")
} else {
    comp_results_800 <- microbenchmark(comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.25, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.25, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.25, method = "exp.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.75, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.75, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 8, pctCat = 0.75, method = "exp.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.25, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.25, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.25, method = "exp.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.75, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.75, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 24, pctCat = 0.75, method = "exp.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.25, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.25, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.25, method = "exp.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.75, method = "marginal"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.75, method = "gamma.frailty"),
                                       comp_time(N = 200, K = 4, nVars = 64, pctCat = 0.75, method = "exp.frailty"),
                                       times = 10)
    save(comp_results_800, file = "comp_results_800.RData")
}

## Sample size 5000
if ("comp_results_5000.RData" %in% list.files()) {
    load("comp_results_5000.RData")
} else {
    comp_results_5000 <- microbenchmark(comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.25, method = "marginal"),
                                        comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.25, method = "gamma.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.25, method = "exp.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.75, method = "marginal"),
                                        comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.75, method = "gamma.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 8, pctCat = 0.75, method = "exp.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.25, method = "marginal"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.25, method = "gamma.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.25, method = "exp.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.75, method = "marginal"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.75, method = "gamma.frailty"),
                                        comp_time(N = 500, K = 10, nVars = 24, pctCat = 0.75, method = "exp.frailty"),
                                        times = 5)
    save(comp_results_5000, file = "comp_results_5000.RData")
}

## Reduce number of evaluations with 64 variables (takes a long time).
if ("comp_results_5000_64vars.RData" %in% list.files()) {
    load("comp_results_5000_64vars.RData")
} else {
    comp_results_5000_64vars <- microbenchmark(comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.25, method = "marginal"),
                                               comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.25, method = "gamma.frailty"),
                                               comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.25, method = "exp.frailty"),
                                               comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.75, method = "marginal"),
                                               comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.75, method = "gamma.frailty"),
                                               comp_time(N = 500, K = 10, nVars = 64, pctCat = 0.75, method = "exp.frailty"),
                                               times = 3)
    save(comp_results_5000_64vars, file = "comp_results_5000_64vars.RData")
}

sumComp <- rbind(summary(comp_results_800), summary(comp_results_5000), summary(comp_results_5000_64vars))
outputData <- data.frame(N = rep(c(200, 500), each = 18), K = rep(c(4, 10), each = 18),
                         nVars = rep(rep(c(8, 24, 64), each = 6), 2),
                         pctCat = rep(rep(c(.25, .75), each = 3), 6),
                         method = rep(c("marginal", "gamma.frailty", "exp.frailty"), 12),
                         neval = sumComp$neval, medianMin = round(sumComp$median/60, 1))
reshape(outputData, timevar = "method",
        idvar = c("N", "K", "nVars", "pctCat", "neval"), direction = "wide")

